@@ -1,3 +1,4 @@
+import asyncio
 import time
 from datetime import datetime
 from typing import Any, Dict, List, Optional
@@ -5,12 +6,12 @@ from uuid import uuid4
 
 import pytz
 from galileo_core.helpers.execution import async_run
+from galileo_core.schemas.shared.workflows.node_type import NodeType
 
 from galileo_observe.schema.transaction import (
     TransactionLoggingMethod,
     TransactionRecord,
     TransactionRecordBatch,
-    TransactionRecordType,
 )
 from galileo_observe.utils.api_client import ApiClient
 
@@ -87,7 +88,7 @@ class GalileoObserve:
 
     def log_node_start(
         self,
-        node_type: TransactionRecordType,
+        node_type: NodeType,
         input_text: str,
         model: Optional[str] = None,
         temperature: Optional[float] = None,
@@ -99,7 +100,7 @@ class GalileoObserve:
 
         Parameters
         ----------
-        node_type : TransactionRecordType
+        node_type : NodeType
             Type of node ("llm", "chat", "chain", "agent", "tool", "retriever")
         input_text : str
             Input to the node as a str or json dump, by default None
@@ -158,7 +159,7 @@ class GalileoObserve:
         node_id : str
             Output value from log_node_start()
         output_text : str
-            Ouput from the node as str or json dump (List[str] for retrievers)
+            Output from the node as str or json dump (List[str] for retrievers)
         num_input_tokens : Optional[int], optional
             Number of input tokens for llm or chat nodes, by default 0
         num_output_tokens : Optional[int], optional
@@ -185,6 +186,50 @@ class GalileoObserve:
 
         self._finalize_node(TransactionRecord(**model_dict))
 
+    async def async_log_node_completion(
+        self,
+        node_id: str,
+        output_text: str,
+        num_input_tokens: Optional[int] = 0,
+        num_output_tokens: Optional[int] = 0,
+        num_total_tokens: Optional[int] = 0,
+        finish_reason: Optional[str] = None,
+        status_code: Optional[int] = 200,
+    ) -> None:
+        """Async log node completion.
+
+        Parameters
+        ----------
+        node_id : str
+            Output value from log_node_start()
+        output_text : str
+            Output from the node as str or json dump (List[str] for retrievers)
+        num_input_tokens : Optional[int], optional
+            Number of input tokens for llm or chat nodes, by default 0
+        num_output_tokens : Optional[int], optional
+            Number of output tokens for llm or chat nodes, by default 0
+        num_total_tokens : Optional[int], optional
+            Total number of tokens for llm or chat nodes, by default 0
+        finish_reason : Optional[str], optional
+            Finish reason for node (e.g. "chain end" or "stop"), by default None
+        status_code : Optional[int], optional
+            HTTP status code for the node, by default 200
+        """
+        latency_ms = await asyncio.to_thread(self._end_node, node_id)
+
+        model_dict = self.records[node_id].model_dump()
+        model_dict.update(
+            output_text=output_text,
+            num_input_tokens=num_input_tokens,
+            num_output_tokens=num_output_tokens,
+            num_total_tokens=num_total_tokens,
+            finish_reason=finish_reason,
+            latency_ms=latency_ms,
+            status_code=status_code,
+        )
+
+        await asyncio.to_thread(self._finalize_node, TransactionRecord(**model_dict))
+
     def log_node_error(
         self, node_id: str, error_message: str, status_code: Optional[int] = 500
     ) -> None:
@@ -193,7 +238,7 @@ class GalileoObserve:
         Parameters
         ----------
         node_id : str
-            Ouput from log_node_start()
+            Output from log_node_start()
         error_message : str
             The error message from the remote system or local application
         status_code : Optional[int], optional
