@@ -21,7 +21,9 @@ class ApiClient:
             try:
                 project = self.get_project_by_name(project_name)
                 if project["type"] not in ["llm_monitor", "galileo_observe"]:
-                    raise Exception(f"Project {project_name} is not a Galileo Observe project")
+                    raise Exception(
+                        f"Project {project_name} is not a Galileo Observe project"
+                    )
                 self.project_id = project["id"]
             except Exception as e:
                 if "not found" in str(e):
@@ -50,10 +52,16 @@ class ApiClient:
         if username and password:
             return self.username_login(username, password).get("access_token", "")
 
-        raise Exception("GALILEO_API_KEY or GALILEO_USERNAME and GALILEO_PASSWORD must be set")
+        raise Exception(
+            "GALILEO_API_KEY or GALILEO_USERNAME and GALILEO_PASSWORD must be set"
+        )
 
     def healthcheck(self) -> bool:
-        async_run(make_request(RequestMethod.GET, base_url=self.base_url, endpoint=Routes.healthcheck))
+        async_run(
+            make_request(
+                RequestMethod.GET, base_url=self.base_url, endpoint=Routes.healthcheck
+            )
+        )
         return True
 
     def username_login(self, username: str, password: str) -> Dict[str, str]:
@@ -156,7 +164,9 @@ class ApiClient:
             headers=headers,
         )
 
-    async def ingest_batch(self, transaction_batch: TransactionRecordBatch) -> Dict[str, str]:
+    async def ingest_batch(
+        self, transaction_batch: TransactionRecordBatch
+    ) -> Dict[str, str]:
         transaction_batch.client_version = version("galileo_observe")
         return await self._make_async_request(
             RequestMethod.POST,
